//This file is "inversion.cc"
//Func. definition of class Inversion
//These functions does structural inversion prior to MC simulation
//By Teck Leong Tan
#define SMALL 1.0e-12
#include "inversion.h"

Inversion::Inversion()
{ 
}
Inversion::Inversion (int num) 
{
  num_types=num;
  ensemble='c';
}

void Inversion::invert_matrix(vector< vector<double> > &a)
{
  int nrow=a.size();
  vector<double> dummy;
  dummy.push_back(0);
  if (a.size()!=a[0].size())
  {
    cout<<"Error in inversion.cc: Not a square matrix."<<endl;
    exit(1);
  }
  B.clear();    
  for (int i=0; i< a[0].size();i++)
  {
    B.push_back(dummy);
  }
  gaussj(a, B);
}


void Inversion::lfit()
{
  //Given a set of data points x [0...ndat-1], E[0...ndat-1] with individual standard deviations 
  //sig[O. .ndat-1], use X2 (chi-sq) minimization to fit for some or all of the coefficients 
  //J[O. .ma-1] of a function that depends linearly on J, E = SUM(J_i * afunc_i(X)). The boolean input 
  //array ia [0...ma-1] indicates by true(1) entries those components of a that should be fitted for, 
  //and by false(0) entries those components that should be held fixed at their input values. 
  //The program returns values for J[O. .ma-1] , X2 = chisq, and the covariance matrix 
  //covar[O. .ma-1] [0. .ma-1]. 
  //(Parameters held fixed will return zero covariances.) The user supplies a routine funcs(x,afunc) 
  //that returns the ma basis functions evaluated at x = x in the array afunc[O. .ma-1].

  int i,j,k,l,m,mfit=0; 
  double ym,wt,sum,sig2i;
  vector<double> dummy; 

  int ndat=E.size();
  int ma=J.size();
  cout<<"ndat= "<<ndat<<" ma= "<<ma<<endl; 
  vector<double> afunc(ma); 
  dummy.push_back(1);
  vector< vector<double> > beta(ma,dummy); 
  cout<<"begin lfit()"<<endl;

  for (j=0;j<ma;j++) 
    if (ia[j]) mfit++; 

  if (mfit == 0) cerr<<"lfit: no parameters to be fitted"; 
  for (j=0;j<mfit;j++) 
  { 
    //Initialize the (symmetric) matrix. 
    for (k=0;k<mfit;k++) covar[j][k]=0.0; 
    beta[j][0]=0.0;
  }

  for (i=0;i<ndat;i++ ) 
  { //Loop over data to accumulate coefficients of normal equations. 
    afunc=phi[i];
    ym=E[i]; 
    if (mfit < ma) 
    { //Subtract off dependences on known pieces of the fitting function. 
      for (j=0;j<ma;j++) 
        if (!ia[j]) ym -= J[j]*afunc[j] ; 
    } 
    sig2i=1.0/(sig[i]*sig[i]); 
    for (j=0,l=0;l<ma;l++) 
    { 
      if(ia[l]) 
      { 
        wt=afunc[l]*sig2i; 
        for (k=0, m=0;m<=l;m++) 
          if (ia[m]) covar[j][k++ ]+=wt*afunc[m]; 
        beta[j++][0] += ym*wt;
      }
    }
  }

  for (j=1;j<mfit;j++) //Fill in above the diagonal from symmetry. 
    for (k=0;k<j;k++) 
      covar[k][j]=covar[j][k]; 

  vector< vector<double> > temp; 
  //vector<double> dummy;
  dummy.assign(mfit,0);
  temp.assign(mfit, dummy);//Initialise temp
  cout<<"temp.size()= "<<temp.size()<<endl;
  for (j=0;j<mfit;j++) 
    for (k=0;k<mfit;k++) 
      temp [j][k]=covar[j][k]; 

  //Checking
  /*cout<<"temp"<<endl;
  for (int i=0; i<temp.size();i++)
  {
    for (int j=0;j<temp[i].size();j++)
    {
      cout<<temp[i][j]<<" ";
    }
    cout<<endl;
  }
  cout<<endl;*/

  //Checking
  /*cout<<"beta"<<endl;
  for (int i=0; i<beta.size();i++)
  {
    for (int j=0;j<beta[i].size();j++)
    {
      cout<<beta[i][j]<<" ";
    }
    cout<<endl;
  }
  cout<<endl;*/

  issingular=0;
  issingular=gaussj (temp, beta); //Matrix inversion.
     //if(issingular==1) then CV=inf;
 
  cout<<"after gaussj()"<<endl; 
  for (j=0;j<mfit;j++) 
    for (k=0;k<mfit;k++) 
      covar[j][k]=temp[j][k]; 

  for (j=0,l=0;l<ma;l++) 
    if (ia[l]) J[l]=beta[j++][0];//Partition solution to appropriate coefficients 
    chisq=0.0; 
    double tempd;
    for (i=0;i<ndat;i++) 
    { 
      //Evalulate chi-square fit
      afunc=phi[i];
      sum=0.0; 
      for(j=0;j<ma;j++) sum+= J[j]*afunc[j]; 
      tempd=(E[i]-sum)/sig[i];
      chisq += tempd*tempd; 
    }

  covsrt(ia,mfit);//Sort covariance matrix to orginal order of fitting coefficients
}

void Inversion::covsrt(vector<bool> &ia, const int mfit) 
{//Expand in storage the covariance matrix covar, so as to 
 //take into account parameters that are being held fixed. 
 //(For the latter, return zero covariances.) 
 int i,j,k; 
 int ma=ia.size(); 
 for (i=mfit;i<ma;i++) 
   for (j=0;j<i+1;j++) covar[i][j]=covar[j][i]=0.0; 
 k=mfit-1; 
 for (j=ma-1;j>=0;j--) 
 { 
   if (ia[j]) 
   {
     for(i=0;i<ma;i++) swap(covar[i][k],covar[i][j]); 
     for(i=0;i<ma;i++) swap(covar[k][i],covar[j][i]); 
     k--;
   }
 }
}

int Inversion::gaussj(vector< vector<double> > &a, vector< vector<double> >&b)
{
  int i,icol,irow,j,k,l,ll; 
  double big, dum, pivinv;

  int n=a.size();//number of rows in a 
  int m=b[0].size();//number of vector columns in b
  vector<int> indxc(n,0), indxr(n,0), ipiv(n,0) ;//for bookeeping 
  
  //for (j=0;j<n;j++) ipiv[j]=0;
  for (i=0;i<n;i++) 
  { 
    big=0.0; 
    for (j=0;j<n;j++)
    { 
      if (ipiv[j]!= 1)
      { 
        for (k=0;k<n;k++) 
        { 
          if (ipiv[k] == 0) 
          { 
            if (fabs(a[j][k]) >= big) 
            { 
              big=fabs(a[j][k]); 
              irow=j; 
              icol=k; 
            }
          }
        }
      }
    }
    ++(ipiv[icol]);;//Obtained pivot element

    //Row interchanging. Columns are not interchanged, only relabled.
    if (irow!=icol) 
    { 
      for (l=0;l<n;l++) {swap(a[irow][l],a[icol][l]);} 
      for (l=0;l<m;l++) {swap(b[irow][l],b[icol][l]);} 
     }
    indxr[i]=irow;
    indxc[i]=icol;
    if (a[icol][icol] == 0.0){
      cerr<<"gaussj: Singular Matrix"<<endl; 
      return 1;
    }
    if (fabs(a[icol][icol]) < SMALL) 
    {
      cerr<<"gaussj: Suspected Singular Matrix"<<endl;
      return 1;
    }
    pivinv=1.0/a[icol][icol]; 
    a[icol][icol]=1.0; 
    for (l=0;l<n;l++) a[icol][l]*= pivinv; 
    for (l=0;l<m;l++) b[icol][l]*= pivinv; 
 
    //Row reduction
    for (ll=0;ll<n;ll++) 
      if (ll != icol)
      {
        dum=a[ll][icol];
        a[ll][icol]=0.0; 
        for (l=0;l<n;l++) a[ll][l] -= a[icol][l]*dum; 
        for (l=0;l<m;l++) b[ll][l] -= b[icol][l]*dum;
      }
  }
   
  //Unscrambling effect of column interchanges  
  for (l=n-l;l>=0;l--) 
  { 
    if (indxr[l] != indxc[l]) 
      for (k=0;k<n;k++) 
        swap(a[k][indxr[l]],a[k][indxc[l]]); 
  }
  
  return 0;
}//end of gaussj()

void Inversion::swap(double &a, double& b)
{
  double temp;
  temp=a;
  a=b;
  b=temp;
}

void Inversion::set_param()
{
  //E=phi*J  where E is set of energy data. phi is correlation matrix. 
  //J interaction energy (to be solved for).  

  //Initialise covar matrix
  vector<double> dummy2;
  dummy2.assign(J.size(),0);
  covar.assign(J.size(),dummy2);
  sig.assign(E.size(),1);//Sigma is set to 1.
  ia.assign(J.size(),1);//All elements of J aren't fixed
  if (ensemble=='c')//if fixed composition
  {
    //All 1-body interactions will be not be fitted for
    for (int i=0;i<num_types;i++)
    {
      ia[i+1]=0;
    }  
  }  
      
  //Do a check:
  if(phi.size()<phi[0].size())
  {
    cout<<"Error in Inversion::set_param(): No. of data sets should not ";
    cout<<"be less than the no. of paramters to be fitted."<<endl;
    //exit(1);
  }
}

void Inversion::cal_cv1(const vector<int> &nc, const ClusterTree &clustree)
{
  //E=phi*J  where E is set of energy data. phi is correlation matrix. 
  //J interaction energy (to be solved for).  
                                                                        
  //Initialise covar matrix
  vector<double> dummy2;
  dummy2.assign(J.size(),0);
  covar.assign(J.size(),dummy2);
  sig.assign(E.size(),1);//Sigma is set to 1.
  ia.assign(J.size(),0);//Initialise all J elements as fixed (i.e not to be fitted)
  ia[0]=1;//Constant term will always be fitted for
  if (ensemble!='c')//if not fixed composition
  {
    //All 1-body interactions will be be fitted for
    for (int i=0;i<num_types;i++)
    {
      ia[i+1]=1;
    }  
  }  
  //Let's set ia appropriately
  for (int n=2;n<nc.size();n++)
  {
    int startind=clustree.sublat[0].nbody[n].all_clus[0][0];
    int endind=startind+clustree.sublat[0].nbody[n].all_clus[nc[n]][power(num_types,n)-1];
    //Update ia
    for (int i=startind;i<=endind;i++)
    {
      ia[i]=1;
    }  
  }  
  
  //Do a check:
  if(phi.size()<phi[0].size())
  {
    cout<<"Error in Inversion::cal_cv1(): No. of data sets should not ";
    cout<<"be less than the no. of paramters to be fitted."<<endl;
    //exit(1);
  }

  //Now let's do lfit and claculate cv-1 score
  lfit(); //Structural inversion with general least squares.
  //cout<<"lfit() done"<<endl;
  cv1_fast();
}	  

void Inversion::print_outphi()
{
  //Print to file phi and E
  fstream outphi;
  outphi.open("outphi.txt", ios::out);
  if (!outphi)
  {
    cout<<"File: "<<"outphi.txt"<<" is not opened!"<<endl;
    exit(1);
  }
  outphi<<"phi :" <<endl;
  for (int i=0; i<phi.size();i++)
  {
    outphi<<E[i]<<"  ";
    for (int j=0;j<phi[i].size();j++)
    {
      outphi<<setw(10)<<setprecision(6)<<phi[i][j]<<" ";
    }
    
    outphi<<"   sid: "<<sid[i];
    outphi<<endl;
  }
  outphi<<endl;
  
  outphi<<"ia: "<<endl;
  for (int i=0; i<ia.size();i++)
  {
    outphi<<ia[i]<<" ";
  }
  outphi<<endl<<endl;

  outphi<<"E vector"<<endl;
  for (int i=0; i<E.size();i++)
  {
    outphi<<E[i]<<endl;
  }
  outphi<<endl<<endl;
  outphi.close();
  //print_matrix("out_inistruct");
}

 
void Inversion::print_matrix(char* out_structfile )
{
  fstream out_struct;
  out_struct.open(out_structfile, ios::out);
  if (!out_struct)
  {
    cout<<"File: "<<out_structfile<<" is not opened!"<<endl;
    exit(1);
  }

  double ls=0;//least square
  double Ece;//Energy calculated by CE
  //Print out table
  out_struct<<"#sid "<<" Edft ";
  out_struct<<" Ece "<<"  |Edft-Ece| "<<" Composition "<<endl;
  for (int i=0;i<E.size();i++)
  {
    Ece=0;
    for(int j=0;j<phi[i].size();j++)
    {
      Ece+=(J[j]*phi[i][j]);
    }
    out_struct<<left<<setw(10)<<sid[i];
    out_struct<<setprecision(6)<<fixed<<"  "<<E[i]<<"  "<<Ece<<"  ";
    out_struct<<fabs(E[i]-Ece)<<"  ";
    for (int k=1;k<num_types+1;k++)
    {
        out_struct<<phi[i][k]<<"  ";
    }
    out_struct<<endl;
    ls+=((Ece-E[i])*(Ece-E[i]));
  }
  //out_struct<<"Least sqaure= "<<ls/E.size()<<endl;
  cout<<endl;
  if (ensemble=='c')
  {
    out_struct<<"E-disorder= "<<Edisorder<<endl;
  }
  //int arow, acol, brow, bcol;
  //in_struct>>arow>>acol;
  out_struct<<endl<<"Covariance matrix"<<endl;
  for (int i=0; i<covar.size();i++)
  {
    for (int j=0;j<covar[i].size();j++)
    {
      out_struct<<setprecision(6)<<setw(12)<<right<<fixed<<covar[i][j]<<" ";
    }
    out_struct<<endl;
  }
  out_struct<<endl;

  out_struct<<"J vector"<<endl;
  for (int i=0; i<J.size();i++)
  {
    out_struct<<J[i]<<endl;
  }
  out_struct<<endl<<endl;

  out_struct<<"L.S scores: "<<endl;
  out_struct<<sqrt(chisq/sid.size())<<endl<<endl;
  out_struct<<"CV-1 scores: "<<endl;
  for (int i=0; i<cv1_score.size();i++)
  {
      out_struct<<sqrt(cv1_score[i])<<endl;
  }

  out_struct.close();
}

double Inversion::dot_product(const vector<double> &a, const vector<double>& b)
{
  int size =a.size();
  if (size!=b.size())
  {cerr<<"Error in inversion.cc: Vector sizes must be the same for dot product"<<endl; exit(1);}
  double dot_prod=0;
  for (int k=0;k<size;k++)
  {
    dot_prod+=(a[k]*b[k]);
  }
  return dot_prod;
}  







